package  nstat.patch;

import   juice.util.*;
import    jam.operator.*;
import    rsf.vector.*;
import   java.util.Enumeration;
import   java.util.NoSuchElementException;

/** 
 * The Patcher enumerates all patches given a quilt vector, the patch size, 
 * and the desired overlap. The Patcher object is best created with the 
 * PatcherFactory that computes the precise overlap values and number of 
 * patches to fit the patches into the quilt space. The Patcher encapsulates 
 * a Window operator.
 *
 * <BR><!--#include file="./source.inc"--><BR>
 * 
 * @see PatcherFactory
 * @see PatcherTest 
 * @see Window
 * @see PatchPushOp
 */
public class Patcher implements Enumeration {
  protected RsfSpace  quiltSpace;
  protected int[]     patchSize;
  protected int[]     nPatches;
  /** The actual overlap which is distinct from the desired overlap*/
  protected float[]   overlap;
  // Enumeration
  int[]   count; 
  boolean exhausted = false; 

  /** 
   * Should only be used by PatcherFactory since it requires the actual 
   * overlap values not the desired overlap. <br>
   */
  Patcher(RsfSpace quiltSpace,int[] patchSize,int[] nPatches,float[] overlap) {
    this.quiltSpace = quiltSpace;
    this.patchSize  = patchSize;
    this.nPatches   = nPatches;
    this.overlap    = overlap;
    count           = new int[this.nPatches.length]; // zero by default!
   }

  /** @return the total number of patches in the quilt */
  public int getNPatches() {
    int n = 1;
    for (int i = 0; i < nPatches.length; i++)
      n *= nPatches[i];
    return n;
  }
  /** @return the number of patches in a particular dimension */
  public int getNPatches(int dim) { return nPatches[dim]; }

  /** Better name might be getWindowOp().*/ 
  public hasAdjoint getWindow (int[] patchIndex) {
    return new WindowFactory(quiltSpace, getAnchor(patchIndex),
			     patchSize).getForward();
  }

  int[] getAnchor(int[] patchIndex) {
    int n = quiltSpace.getDimension();
    if (n != patchIndex.length) {
      System.out.println("Error: dimensions of quilt space and patch " +
			 "index aren't equal");
      System.out.println("dim(quiltSpace) = " + n);
      System.out.println("dim(patchIndex) = " + patchIndex.length);
      return null;
    }
    int[] anchor = new int[n];
    for (int i = 0; i < n; i++)
      anchor[i] = get1Anchor(i, patchIndex[i]);
    return anchor;
  }

  int get1Anchor(int dim, int patchIndex) {
    return (int) ((1F - overlap[dim]) * (patchIndex * patchSize[dim])
		  + 0.5F);
  }

  //Enumeration 

  public boolean hasMoreElements() {return ! this.exhausted;}
  public boolean hasMorePatches () {return ! this.exhausted;}
  /** 
   * convenient when the non-stationary operation has identical domain and 
   * range. 
   * @return cohy.window.Window operator
   */
  public Object  nextElement() {
    synchronized (this) {
      if (!exhausted) {
	hasAdjoint p = this.getWindow(count);
	incrementCount();
	return p;
      }
    }
    throw new NoSuchElementException("PatchEnumerator");
  }
  /** 
   * Assumes that quilt is in quiltSpace. 
   * @return The current patch extracted from the quilt. 
   */ 
  public Rsf nextPatch(Rsf quilt) {
    Window cut = (Window) this.nextElement();
    return (Rsf) cut.image(quilt);
  }

  /**
   * Increment count[] according to patchSize[].
   * @return false if incrementing count brings it back to zero, true otherwise
   */
  void incrementCount() {
    for (int i = 0; i < count.length; i++) {
      count[i]++;
      if (count[i] < this.nPatches[i])
	return;
      count[i] = 0;
    }
    exhausted = true;
  }
  public String toString() {
    String val = 
      "Patcher:   \n" + 
      "  quiltSzie: "+ 
                   JuiceUtil.arrayToString(quiltSpace.getNSamples(),",")+"\n"+
      "  patchSzie: "+JuiceUtil.arrayToString(patchSize,            ",")+"\n"+
      "  nPatches:  "+JuiceUtil.arrayToString(nPatches,             ",")+"\n"+
      "  overlap:   "+JuiceUtil.arrayToString(overlap,              ","); 
    return val; 
  }
}

