package  nstat.patch;

import    jam.operator.*;
import    rsf.vector.*;
import   java.util.Enumeration;
import   java.util.NoSuchElementException;

/**
 * The factory constructs a Patcher object with optimal overlap and 
 * and number of patches. The overlap values approximate the desired
 * values stated at construction of the factory. 
 * Can this class be static?
 *
 * BUG: inconsistent argument order: 
 * patchSize overlap or overlap patchSize. Search.
 *
 * @see Patcher
 * @see Window 
 * @see PatchPushOp
 * 
 * <BR><!--#include file="./source.inc"--><BR>
 */
public class PatcherFactory {
  protected RsfSpace quiltSpace;
  protected int[]    patchSize;
  protected int[]    nPatches;
  protected float[]  overlap;

  /**
   * Defaults to identical 0.5 overlap value for all dimensions.
   */
  public PatcherFactory(RsfSpace quiltSpace, int[] patchSize) {
    this(quiltSpace, patchSize, 0.5F);
  }
  /**
   * Identical overlap value for all dimensions
   */
  public PatcherFactory(RsfSpace quiltSpace, int[] patchSize, float overlap){
    this(quiltSpace, patchSize, 
	 juice.util.JuiceUtil.arrayfill(patchSize.length, overlap));
  }
  /**
   * @param overlap    desired fraction of patch overlap [0-1]. 
   *                   Negative overlap causes gaps between patches. 
   * @param patchSize  number of Samples for each patch axis. 
   * @param quiltSpace 
   */
  public PatcherFactory(RsfSpace quiltSpace, int[] patchSize, float[] overlap){
    int n = patchSize.length;
    if (quiltSpace.getDimension() != n) {
      System.err.println("Error in PatcherFactory constructor: dimensions of "+
			 "quiltSpace and patchSize are unequal");
      System.err.println("dim(quiltSpace) = " + quiltSpace.getDimension());
      System.err.println("dim(patchSize ) = " + n);
      return;
    }
    if (overlap.length != n) {
      System.err.println("Error in PatcherFactory constructor: dimensions of "+
			 "overlap and patchSize are unequal");
      System.err.println("dim(overlap)     = " + overlap.length);
      System.err.println("dim(patchSize ) = " + n);
      return;
    }
    this.quiltSpace = quiltSpace;
    this.overlap   = new float[n];
    this.nPatches  = new   int[n];
    this.patchSize  = new   int[n];
    System.arraycopy(patchSize, 0, this.patchSize, 0, n);
    for (int i = 0; i < n; i++)
      setOverlap(i, overlap[i]);
  }

  /** 
   * @return Patcher object that enumerates the defined patches. 
   */ 
  public Patcher getPatcher() {
    return new Patcher(quiltSpace, patchSize, nPatches, overlap);
  }

  static int getNPatches(int   quiltSize,
			   int   patchSize,
			   float desiredOverlap) {
    int n = getLowerNPatches(quiltSize, patchSize, desiredOverlap);
    float loOverlap = getFractionalOverlap(quiltSize,patchSize,n);
    float hiOverlap = getFractionalOverlap(quiltSize,patchSize,n+1);
    float hiErr = Math.abs(hiOverlap - desiredOverlap);
    float loErr = Math.abs(loOverlap - desiredOverlap);
    return (loErr <= hiErr) ? n:(n+1);
  }

  /**
   * @param  n numberOfPatches 
   * @return n * patchSize - quiltSize
   */
  static int getTotalOverlap(int quiltSize ,
			     int patchSize ,
			     int n) {
    return (n * patchSize - quiltSize);
  }

  static float getFractionalOverlap(int quiltSize,
				    int patchSize,
				    int nPatches) {
    if (nPatches == 1) return 0F;
    return (getTotalOverlap(quiltSize, patchSize, nPatches)
	    / ((float) (nPatches-1)*patchSize));
  }

  /**
   * @param desiredOverlap the fraction (between 0,1) of patch overlap desired
   * @return N, where N or N+1 is the corresponding number of patches to use.
   */
  static int getLowerNPatches(int quiltSize,
				int patchSize,
				float desiredOverlap) {
    if (quiltSize == patchSize) 
      return 1;
    else { 
      float margin = (1F - desiredOverlap) * patchSize;
      return Math.max(2, (int) ((quiltSize - patchSize) / margin));
    }
  }

  /**
   * setOverlap sets the overlap for a particular dimension. 
   * The overlap is a desired overlap and will be approximated. 
   * @param dim            the specific dimension for which to set the overlap
   * @param desiredOverlap the fractional patch overlap desired.
   */
  public void  setOverlap(int dim, float desiredOverlap) {
//    if ((desiredOverlap < 0F) || (desiredOverlap >= 1F)) {
    if ((desiredOverlap >= 1F)) {
      System.err.println("Error: desired overlap must be between 0 and 1 " +
			 "(exclusive).");
      return;
    }
    int qSize     = quiltSpace.getNSamples(dim);
    nPatches[dim] = getNPatches(qSize, patchSize[dim], desiredOverlap);
    overlap [dim] = getFractionalOverlap(qSize, patchSize[dim], nPatches[dim]);
  }
  /** @return the approximated overlap for the given dimension */ 
  public float getOverlap(int dim) { return overlap[dim]; }
}

