package  nstat.patch;

import    sep.operator.SepLinearOperator;
import 	  jam.vector.*;
import 	  jam.operator.*;
import    jam.JamObject;
import  juice.util.JuiceUtil;
import 	  rsf.vector.*;

/**
 * Window implements the Truncation/Zero-Padding operator on Rsf's.<BR>
 * The sense of the operator is:<BR>
 * forward = truncation         <BR>
 * adjoint = zero padding       <BR><BR>
 * 
 * Needs a common interface with other Window operations:       <BR>
 * Window utility (iterator, dim reduction, midpoint and width).<BR>
 * out-of-core Rsf.                                             <BR>
 * 
 * @see WindowFactory
 * @see WindowTest 
 * @see Patcher
 * 
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @author: Matt Schwab and Joel Schroeder
 */
public class Window extends SepLinearOperator {
  protected int[]    anchor;
  protected int      dim;
  protected RsfSpace wallSpace;
  protected RsfSpace windSpace;
  protected Rsf      windTent;

  /**
   * You probably never want to invoke this, since the windSpace and the 
   * anchorPoint have to be in agreement. If they are not then the 
   * anchor succeeds. 
   * Use WindowFactory or WindowEnumeration instead.<Br>
   * 
   * Window of length 1 work but not thoroughly tested. 
   * Careful: a 3-D Rsf with only 1 sample in one axis is different from 
   * a 2-D Rsf.  
   *
   * @param wallSpace  the superspace which is (much) larger than the window
   * @param windSpace  the window space (subset of wallSpace).
   * @param anchorPoint the lowest-index vertex of window on the data wall 
   *        (e.g. the lower-left corner in Cartesian 2D coordinates)
   * @param isFwd       <TT>true</TT>  is for a truncation   operator <BR>
   *                    <TT>false</TT> is for a zero-padding operator
   */
  public Window(RsfSpace wallSpace,
	        RsfSpace windSpace,
		int[]    anchorPoint, 
	        boolean  isFwd ) {
    super(wallSpace, windSpace, isFwd);
    if (isFwd) {
      this.wallSpace = (RsfSpace) this.getDomain();
      this.windSpace = (RsfSpace) this.getRange();
    }
    else {
      this.wallSpace = (RsfSpace) this.getRange();
      this.windSpace = (RsfSpace) this.getDomain();
    }
    dim = wallSpace.getDimension();
    if (dim != windSpace.getDimension()) 
      System.err.println("Dimension of wall and window different");
    setAnchor (anchorPoint);
  }
  /** casts to constructor for RsfSpaces */ 
  public Window(Space wallSpace,
	        Space windSpace, int[] anchorPoint, boolean isFwd) {
    this((RsfSpace)   wallSpace,
	 (RsfSpace)   windSpace,       anchorPoint,         isFwd);
  }

  /** copy constructor */
  protected Window(Window src) {
    super(src);
    this.dim    = src.dim;
    this.anchor = src.getAnchor();
  }

  /** 
   * @return a copy of this window's anchor 
   */
  public    int[] getAnchor() {
    int[] anch = new int[dim];
    System.arraycopy(anchor, 0, anch, 0, dim);
    return anch;
  }

  /** 
   * Warning! 
   * anchorPoint changes modify the range  ( isFwd) Truncation 
   *              or                domain (!isFwd) Zero-Padding 
   */
  public void  setAnchor(int[] anchorPoint) {
    if (dim != anchorPoint.length) {
      System.err.println("Error in setAnchor(): lengths of arrays unequal.");
      return;
    }
    if (this.anchor == null) this.anchor = new int[dim];
    System.arraycopy(anchorPoint, 0, anchor, 0, anchorPoint.length);
    for (int i=0; i< this.dim; i++) {
      this.windSpace.getAxes()[i].offset = 
      this.wallSpace.getAxes()[i].offset + anchor[i] * 
      this.wallSpace.getAxes()[i].delta;
    }
  }

  protected void apply(boolean add, Vector qq, Vector pp) {
    if (!add) {
      if (isFwd) pp.zero();
      else       qq.zero();
    }
    switch (dim) {
    case   1:
      apply1D(add, (Rsf1) qq, (Rsf1) pp);
      break;
    case   2:
      apply2D(add, (Rsf2) qq, (Rsf2) pp);
      break;
    case   3:
      apply3D(add, (Rsf3) qq, (Rsf3) pp);
      break;
    default:
	System.err.println("Error: Window only supported" +
			   " up to " + 3 + " dimensions");
    }
  }

  void apply1D(boolean add, Rsf1 q, Rsf1 p) {
    float[] qq = q.d, pp = p.d;
    int j;
    for (int i = 0; i < pp.length; i++) { j = i + anchor[0];
      if (isFwd)
	pp[i] += qq[j];
      else
	qq[j] += pp[i];
    }
  }

  void apply2D( boolean add, Rsf2 q, Rsf2 p) {
    float[][] qq = q.d, pp = p.d;
    int  []   n  = ((RsfSpace) p.getSpace()).getNSamples();
    for (int p1 = 0; p1 < n[1]; p1++) { int q1 = p1 + anchor[1];
    for (int p0 = 0; p0 < n[0]; p0++) { int q0 = p0 + anchor[0];
	if (isFwd)
	  pp[p1][p0] += qq[q1][q0];
	else
	  qq[q1][q0] += pp[p1][p0];
    }}
  }

  void apply3D( boolean add, Rsf3 q, Rsf3 p) {
    float[][][] qq = q.d, pp = p.d;
    int  []     n  = ((RsfSpace) p.getSpace()).getNSamples();
    for (int p2 = 0; p2 < n[2]; p2++) { int q2 = p2 + anchor[2];
    for (int p1 = 0; p1 < n[1]; p1++) { int q1 = p1 + anchor[1];
    for (int p0 = 0; p0 < n[0]; p0++) { int q0 = p0 + anchor[0];
	if (isFwd)
	  pp[p2][p1][p0] += qq[q2][q1][q0];
	else
	  qq[q2][q1][q0] += pp[p2][p1][p0];
    }}}
  }

  public boolean equals(Object obj2) {
    if (this == obj2)  return true;
    boolean equal = true; 
    if (! (obj2 instanceof Window))                        equal = false;
    Window pat2 = (Window) obj2;
    if (! super.equals(pat2))                              equal = false;
    if (this.dim != pat2.dim)                              equal = false;
    if (! JuiceUtil.equalArrays(this.anchor, pat2.anchor)) equal = false;
    return equal; 
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("Window Operator (");
    sb.append((isFwd ? "truncation":"zero-padding") + ")\n");
    sb.append("Quilt Space :\n" +  input);
    sb.append("Window Space :\n" + output);
    sb.append("Anchor point:\n" + JuiceUtil.arrayToString(anchor, ", "));
    sb.append("\n");
    return sb.toString();
  }

  public JamObject copy() { return new Window(this); }
}




