package  nstat.patch;

import    sep.operator.SepLinearOperator;
import 	  jam.vector.*;
import 	  jam.operator.*;
import 	  rsf.vector.*;

/**
 * WindowFactory creates a Window operator for a given domain and/or range.<BR>
 * The factory ensures the consistency of the domain, range, and anchor 
 * specifications of the Window operator. 
 * 
 * @see Window
 * @see WindowTest 
 * @see Patcher 
 * 
 * Warning: currently does not implement the Factory interface. 
 * 
 * <BR><!--#include file="./source.inc"--><BR>
 */
public class WindowFactory {
  protected Window fwd;

  /**
   * @param wallSpace the space of the Rsf from which a window is taken
   * @param anchor     the index into the wallSpace of the "smallest" vertex
   * @param windSz  the dimensions of the window
   */
  public WindowFactory(RsfSpace wallSpace,
		      int[]    anchor    ,
		      int[]    windSz ) {
    RsfSpace windowSpace = createWindowSpace(wallSpace, anchor, windSz);
    fwd = new Window(wallSpace, windowSpace, anchor, true );
  }

  /** 
   * if windSpace is not an exact subspace of wallSpace, this will create a 
   * Window operator where windSpace is a subspace of wallSpace 
   * (approximation). However, the Window operator methods do not check the 
   * spaces. This problem arises through limited precision with float values 
   * of the axes' increment, and offset. 
   */
  public WindowFactory(RsfSpace wallSpace,
	               RsfSpace windSpace) {
    int[] anchor = createAnchor(wallSpace, windSpace); 
    int[] windSz = windSpace.getNSamples(); 
    RsfSpace windowSpace = createWindowSpace(wallSpace, anchor, windSz);
    fwd = new Window(wallSpace, windowSpace, anchor, true );
  }

  /** 
   * Does not check the full compatibility of Axes. 
   * Finds the wall sample nearest to the window's offset. 
   */
  int[] createAnchor(RsfSpace wallSpace, RsfSpace windSpace) {
    int n = wallSpace.getDimension();
    if (n != windSpace.getDimension()) { 
      System.err.println("Error: dimensions of wall and window not equal"); 
      return null; 
    }
    Axis[] wallAxes = wallSpace.getAxes();
    Axis[] windAxes = windSpace.getAxes();
    int [] anchor   = new int[n]; 
    for (int i=0; i<n; i++) {
      anchor[i] = 
	(int)(.5f + (windAxes[i].offset-wallAxes[i].offset)/wallAxes[i].delta);
    }
    return anchor; 
  }

  RsfSpace createWindowSpace(RsfSpace wallSpace,
			     int[]    anchor    ,
			     int[]    windSz ) {
    if (wallSpace.getDimension() != windSz.length) 
      System.err.println("Dimensions incompatible ... "); 
    Axis[]  wallAxes = wallSpace.getAxes();
    int n = wallAxes.length;
    Axis[]  windowAxes = new Axis[n];
    for (int i = 0 ; i < n ; i++ ) {
      windowAxes[i]           = new Axis(wallAxes[i]);
      windowAxes[i].nsamples  = windSz[i];
      windowAxes[i].offset   += wallAxes[i].delta * anchor[i];
    }
    return new RsfSpace(windowAxes);
  }

  /**
   * @returns the forward form of the windowing operator, truncation.
   */
  public Window     getForward() { return fwd; }
  /**
   * @returns the adjoint form of the windowing operator, zero-padding.
   */
  public Window     getAdjoint() { return (Window) fwd.adjoint(); }
  //  public hasAdjoint getAdjoint() { return fwd.adjoint(); }
}

