package  nstat.patchF;

import    sep.operator.SepLinearOperator;
import 	  jam.vector.*;
import 	  jam.operator.*;
import    jam.JamObject;
import  juice.util.JuiceUtil;
import 	  rsfF.vector.*;

/**
 * Patch implements the Truncation/Zero-Padding operator on Rsf's.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * The sense of the operator is:<BR>
 * forward = truncation
 * adjoint = zero padding
 *
 */
public class Patch extends SepLinearOperator {
  protected int[] anchor;
  protected int   dim;

  /** The maximum dimension of Rsf supported by this operator. */
  public static int maxDim = 10;

  /**
   * @param quiltSpace  the superspace which is (much) larger than the patch
   * @param patchSpace  the space of the patch
   * @param anchorPoint the lowest-index vertex of patch within quilt (e.g. the lower-left corner in Cartesian 2D coordinates)
   * @param isFwd       <TT>true</TT>  is for a truncation   operator <BR>
   *                    <TT>false</TT> is for a zero-padding operator
   */
  public Patch(RsfSpace quiltSpace,
	       RsfSpace patchSpace,
	       int[]    anchorPoint,
	       boolean  isFwd )
  {
    super(quiltSpace, patchSpace, isFwd);
    
    dim = quiltSpace.getDimension();
    if (dim > maxDim) 
      System.err.println("Error: Patch only supported" +
			 " up to " + maxDim + " dimensions");
    setAnchor (anchorPoint);
  }

  /** copy constructor */
  protected Patch(Patch src) {
    super(src);
    this.dim    = src.dim;
    this.anchor = src.getAnchor();
  }

  /** returns a copy of this patch's anchor */
  public    int[] getAnchor() {
    int[] anch = new int[dim];
    System.arraycopy(anchor, 0, anch, 0, dim);
    return anch;
  }

  /** copies anchorPoint to be the new anchor */
  public    void  setAnchor(int[] anchorPoint) {
    if (this.anchor == null) this.anchor = new int[10];
    if (dim != anchorPoint.length) {
      System.err.println("Error in setAnchor(): lengths of arrays unequal.");
      return;
    }
    System.arraycopy(anchorPoint, 0, anchor, 0, anchorPoint.length);
  }

  protected void apply(boolean add, Vector qq, Vector pp) {
    if (!add) {
      if (isFwd) pp.zero();
      else       qq.zero();
    }
    switch (dim) {
    case   1:
      apply1D( add, (Rsf) qq, (Rsf) pp);
      break;
    case   2:
      apply2D( add, (Rsf) qq, (Rsf) pp);
      break;
    default:
      if (dim > maxDim)
	System.err.println("Error: Patch only supported" +
			   " up to " + maxDim + " dimensions");
      else
	applyManyD( add, (Rsf) qq, (Rsf) pp);
      break;
    }
  }

  void apply1D(boolean add, Rsf q, Rsf p) {
    float[] qq = q.d, pp = p.d;
    int j = anchor[0];

    for (int i = 0; i < pp.length; i++, j++)
      if (isFwd)
	pp[i] += qq[j];
      else
	qq[j] += pp[i];
  }

  void apply2D( boolean add, Rsf q, Rsf p) {
    float[] qq = q.d, pp = p.d;

    int[] qi = new int[2], pi = new int[2];
    int[] n = ((RsfSpace) p.getSpace()).getDimensions();
    System.arraycopy(anchor, 0, qi, 0, 2);
    for ( pi[1] = 0; pi[1] < n[1]; pi[1]++) { qi[1] = pi[1] + anchor[1];
    for ( pi[0] = 0; pi[0] < n[0]; pi[0]++) { qi[0] = pi[0] + anchor[0];
	if (isFwd)
	  pp[p.i(pi)] += qq[q.i(qi)];
	else
	  qq[q.i(qi)] += pp[p.i(pi)];
    }}
  }

  void applyManyD( boolean add, Rsf q, Rsf p) {
    float[] qq = q.d, pp = p.d;

    int[] qi = new int[maxDim];
    int[] pi = new int[maxDim];
    int[]  n = new int[maxDim];
    for (int i = 0; i < maxDim; i++)
      n [i] = ((RsfSpace) p.getSpace()).getNSamples(i);
    System.arraycopy(anchor, 0, qi, 0, maxDim);

    qi[9] = anchor[9];
    for (pi[9] = 0; pi[9] < n[9]; pi[9]++, qi[9]++) { qi[8] = anchor[8];
    for (pi[8] = 0; pi[8] < n[8]; pi[8]++, qi[8]++) { qi[7] = anchor[7];
    for (pi[7] = 0; pi[7] < n[7]; pi[7]++, qi[7]++) { qi[6] = anchor[6];
    for (pi[6] = 0; pi[6] < n[6]; pi[6]++, qi[6]++) { qi[5] = anchor[5];
    for (pi[5] = 0; pi[5] < n[5]; pi[5]++, qi[5]++) { qi[4] = anchor[4];
    for (pi[4] = 0; pi[4] < n[4]; pi[4]++, qi[4]++) { qi[3] = anchor[3];
    for (pi[3] = 0; pi[3] < n[3]; pi[3]++, qi[3]++) { qi[2] = anchor[2];
    for (pi[2] = 0; pi[2] < n[2]; pi[2]++, qi[2]++) { qi[1] = anchor[1];
    for (pi[1] = 0; pi[1] < n[1]; pi[1]++, qi[1]++) { qi[0] = anchor[0];
    for (pi[0] = 0; pi[0] < n[0]; pi[0]++, qi[0]++) {
      if (isFwd)
	pp[p.i(pi)] += qq[q.i(qi)];
      else
	qq[q.i(qi)] += pp[p.i(pi)];
    }}}}}}}}}}
  }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof Patch)) return false;
    Patch pat2 = (Patch) obj2;
    return (super.equals(pat2)   &&
	    this.dim == pat2.dim &&
	    JuiceUtil.equalArrays(this.anchor, pat2.anchor));
  }

  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("Patch Operator (");
    sb.append((isFwd ? "truncation":"zero-padding") + ")\n");
    sb.append("Quilt Space :\n" +  input);
    sb.append("Patch Space :\n" + output);
    sb.append("Anchor point:\n" + JuiceUtil.arrayToString(anchor, ", "));
    sb.append("\n");
    return sb.toString();
  }

  public JamObject copy() { return new Patch(this); }
}
