package  nstat.patchF;

import    jam.operator.*;
import    rsfF.vector.*;
import   java.util.Enumeration;
import   java.util.NoSuchElementException;

/**
 * <BR><!--#include file="./source.inc"--><BR>
 */
public class Patchwork {
  protected RsfSpace quiltSpace;
  protected int[]    patchSize;
  protected int[]    nPatches;
  protected float[]  overlap;

  public Patchwork(RsfSpace quiltSpace, int[] patchSize) {
    this(quiltSpace, patchSize, 0.5F);
  }

  public Patchwork(RsfSpace quiltSpace, int[] patchSize, float overlap) {
    int n = patchSize.length;
    if (quiltSpace.getDimension() != n) {
      System.err.println("Error in Patchwork constructor: dimensions of " +
			 "quiltSpace and patchSize are unequal");
      System.err.println("dim(quiltSpace) = " + quiltSpace.getDimension());
      System.err.println("dim(patchSize ) = " + n);
      return;
    }
    this.quiltSpace = quiltSpace;
    this.overlap    = new float[n];
    this.nPatches   = new   int[n];
    this.patchSize  = new   int[n];
    System.arraycopy(patchSize, 0, this.patchSize, 0, n);
    for (int i = 0; i < n; i++)
      setOverlap(i, overlap);
  }

  static int getNumPatches(int   quiltSize,
			   int   patchSize,
			   float desiredOverlap) {
    int n = getLowerNumPatches(quiltSize, patchSize, desiredOverlap);
    float loOverlap = getFractionalOverlap(quiltSize,patchSize,n);
    float hiOverlap = getFractionalOverlap(quiltSize,patchSize,n+1);
    float hiErr = Math.abs(hiOverlap - desiredOverlap);
    float loErr = Math.abs(loOverlap - desiredOverlap);
    return (loErr < hiErr) ? n:(n+1);
  }

  /** @return the number of patches in the patchwork */
  public int size() {
    int n = 1;
    for (int i = 0; i < nPatches.length; i++)
      n *= nPatches[i];
    return n;
  }
  public int getNPatches(int dim) { return nPatches[dim]; }

  public hasAdjoint getPatch (int[] patchworkIndex) {
    return new PatchFactory(quiltSpace, getAnchor(patchworkIndex),
			    patchSize).getForward();
  }

  public Enumeration getPatches() { return new PatchworkEnumerator(this); }

  int[]        getAnchor(int[] patchworkIndex) {
    int n = quiltSpace.getDimension();
    if (n != patchworkIndex.length) {
      System.out.println("Error: dimensions of quilt space and patchwork " +
			 "index aren't equal");
      System.out.println("dim(quiltSpace)      = " + n);
      System.out.println("dim(patchwork index) = " + patchworkIndex.length);
      return null;
    }
    int[] anchor = new int[n];
    for (int i = 0; i < n; i++)
      anchor[i] = get1Anchor(i, patchworkIndex[i]);
    return anchor;
  }

  int          get1Anchor(int dim, int patchworkIndex) {
    return (int) ((1F - overlap[dim]) * (patchworkIndex * patchSize[dim])
		  + 0.5F);
  }

  /**
   * @return n * patchSize - quiltSize
   */
  static int getTotalOverlap(int quiltSize ,
			     int patchSize ,
			     int n) {
    return (n * patchSize - quiltSize);
  }

  static float getFractionalOverlap(int quiltSize,
				    int patchSize,
				    int numPatches) {
    if (numPatches == 1) return 0F;
    return (getTotalOverlap(quiltSize, patchSize, numPatches)
	    / ((float) (numPatches-1)*patchSize));
  }

  /**
   * @param desiredOverlap the fraction (between 0,1) of patch overlap desired
   * @return N, where N or N+1 is the corresponding number of patches to use.
   */
  static int getLowerNumPatches(int quiltSize,
				int patchSize,
				float desiredOverlap) {
    float margin = (1F - desiredOverlap) * patchSize;
    return Math.max(1, (int) ((quiltSize - patchSize) / margin));
  }

  /**
   * setOverlap attempts to achieve the desiredOverlap in a particular
   * dimension by choosing the appropriate number of patches in that dimension
   * 
   * @param desiredOverlap the fractional (0-1) patch overlap desired
   */
  public void  setOverlap(int dim, float desiredOverlap) {
    if ((desiredOverlap <= 0F) || (desiredOverlap >= 1F)) {
      System.err.println("Error: desired overlap must be between 0 and 1 " +
			 "(exclusive).");
      return;
    }
    int qSize = quiltSpace.getNSamples(dim);
    nPatches[dim] = getNumPatches(qSize, patchSize[dim], desiredOverlap);
    overlap [dim] = getFractionalOverlap(qSize, patchSize[dim], nPatches[dim]);
  }
       
  public float getActualOverlap(int dim) { return overlap[dim]; }

  /*
  public static void doTest(int q, int p, float o) {
    int n = getLowerNumPatches(q,p,o);
    System.out.println("q,p,o = " + q + ", " + p + ", " + o);
    System.out.println("n     = " + n);
    System.out.println("Actual overlap = " +
		       (getTotalOverlap(q,p,n)/(float) ((n-1)*p)));
    System.out.println("Actual overlap (n+1)= " +
		       (getTotalOverlap(q,p,n+1)/(float) (n*p)));
    System.out.println();
  }
  */
}

final
class PatchworkEnumerator implements Enumeration {
  Patchwork patchwork;
  int[]     count;
  boolean   exhausted;
  
  PatchworkEnumerator(Patchwork patchwork) {
    this.patchwork = patchwork;
    count = new int[patchwork.nPatches.length];
    for (int i = 0; i < count.length; i++)
      count[i] = 0;
    exhausted = false;
  }

  /**
   * Increment count[] according to patchSize[].
   *
   * @return false if incrementing count brings it back to zero, true otherwise
   */
  void incrementCount() {
    for (int i = 0; i < count.length; i++) {
      count[i]++;
      if (count[i] < patchwork.nPatches[i])
	return;
      count[i] = 0;
    }
    exhausted = true;
  }

  public boolean hasMoreElements() { return !exhausted; }

  public Object     nextElement () {
    synchronized (patchwork) {
      if (!exhausted) {
	hasAdjoint p = patchwork.getPatch(count);
	incrementCount();
	return p;
      }
    }
    throw new NoSuchElementException("PatchworkEnumerator");
  }
}
